KNOWN_NETWORKS_AND_TOKENS = {
    "base": {
        "chain_id": 8453,
        "tokens": {
            "usdc": {
                "address": "0x833589fCD6eDb6E08f4c7C32D4f71b54bdA02913",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0x60a3E35Cc302bFA44Cb288Bc5a4F316Fdb1adb42",
                "name": "EURC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "base-sepolia": {
        "chain_id": 84532,
        "tokens": {
            "usdc": {
                "address": "0x036CbD53842c5426634e7929541eC2318f3dCF7e",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0x808456652fdb597867f38412077A9182bf77359F",
                "name": "EURC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "avalanche": {
        "chain_id": 43114,
        "tokens": {
            "usdc": {
                "address": "0xB97EF9Ef8734C71904D8002F8b6Bc66Dd9c48a6E",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0xc891eb4cbdeff6e073e859e987815ed1505c2acd",
                "name": "Euro Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "avalanche-fuji": {
        "chain_id": 43113,
        "tokens": {
            "usdc": {
                "address": "0x5425890298aed601595a70AB815c96711a31Bc65",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            },
            "eurc": {
                "address": "0x5E44db7996c682E92a960b65AC713a54AD815c6B",
                "name": "EUR Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "€"
            }
        }
    },
    "sei": {
        "chain_id": 1329,
        "tokens": {
            "usdc": {
                "address": "0xe15fC38F6D8c56aF07bbCBe3BAf5708A2Bf42392",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    "sei-testnet": {
        "chain_id": 1328,
        "tokens": {
            "usdc": {
                "address": "0x4fCF1784B31630811181f670Aea7A7bEF803eaED",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    "polygon": {
        "chain_id": 137,
        "tokens": {
            "usdc": {
                "address": "0x3c499c542cEF5E3811e1192ce70d8cC03d5c3359",
                "name": "USD Coin",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    "polygon-amoy": {
        "chain_id": 80002,
        "tokens": {
            "usdc": {
                "address": "0x41E94Eb019C0762f9Bfcf9Fb1E58725BfB0e7582",
                "name": "USDC",
                "decimals": 6,
                "eip712Version": "2",
                "symbol": "$"
            }
        }
    },
    # These are local development networks only.
    "local": {
        "chain_id": 31337,
        "tokens": {}
    }
}
